@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2019 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2020 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -49,14 +49,6 @@
 #define LIST_INDENT  40
 #define LIST_HEXBIT  18
 
-typedef struct MacroInhibit MacroInhibit;
-
-static struct MacroInhibit {
-    MacroInhibit *next;
-    int level;
-    int inhibiting;
-} *mistack;
-
 static const char xdigit[] = "0123456789ABCDEF";
 
 #define HEX(a,b) (*(a)=xdigit[((b)>>4)&15],(a)[1]=xdigit[(b)&15]);
@@ -135,12 +127,6 @@ static void list_cleanup(void)
     if (!listfp)
         return;
 
-    while (mistack) {
-        MacroInhibit *temp = mistack;
-        mistack = temp->next;
-        nasm_free(temp);
-    }
-
     list_emit();
     fclose(listfp);
     listfp = NULL;
@@ -172,10 +158,6 @@ static void list_init(const char *fname)
     list_errors = NULL;
     listlevel = 0;
     suppress = 0;
-    nasm_new(mistack);
-    mistack->next = NULL;
-    mistack->level = 0;
-    mistack->inhibiting = true;
 }
 
 static void list_out(int64_t offset, char *str)
@@ -276,8 +258,13 @@ static void list_output(const struct out_data *data)
 	break;
     case OUT_RESERVE:
     {
-        if (size)
+        if (size > 8) {
             list_size(offset, "res", size);
+        } else {
+            memset(q, '?', size << 1);
+            q[size << 1] = '\0';
+            list_out(offset, q);
+        }
 	break;
     }
     default:
@@ -287,21 +274,14 @@ static void list_output(const struct out_data *data)
 
 static void list_line(int type, int32_t lineno, const char *line)
 {
+    (void)type;
+
     if (!listfp)
         return;
 
     if (user_nolist)
       return;
 
-    if (mistack && mistack->inhibiting) {
-        if (type == LIST_MACRO)
-            return;
-        else {                  /* pop the m i stack */
-            MacroInhibit *temp = mistack;
-            mistack = temp->next;
-            nasm_free(temp);
-        }
-    }
     list_emit();
     if (lineno >= 0)
         listlineno = lineno;
@@ -311,15 +291,6 @@ static void list_line(int type, int32_t lineno, const char *line)
     listlevel_e = listlevel;
 }
 
-static void mistack_push(bool inhibiting)
-{
-    MacroInhibit *temp = nasm_malloc(sizeof(MacroInhibit));
-    temp->next = mistack;
-    temp->level = listlevel;
-    temp->inhibiting = inhibiting;
-    mistack = temp;
-}
-
 static void list_uplevel(int type, int64_t size)
 {
     if (!listfp)
@@ -338,13 +309,6 @@ static void list_uplevel(int type, int64_t size)
 
     case LIST_INCLUDE:
         listlevel++;
-        if (mistack && mistack->inhibiting)
-            mistack_push(false);
-        break;
-
-    case LIST_MACRO_NOLIST:
-        listlevel++;
-        mistack_push(true);
         break;
 
     default:
@@ -369,16 +333,11 @@ static void list_downlevel(int type)
 
     default:
         listlevel--;
-        while (mistack && mistack->level > listlevel) {
-            MacroInhibit *temp = mistack;
-            mistack = temp->next;
-            nasm_free(temp);
-        }
         break;
     }
 }
 
-static void list_error(errflags severity, const char *fmt, ...)
+static void printf_func(2, 3) list_error(errflags severity, const char *fmt, ...)
 {
     va_list ap;
 
