@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
- *   
- *   Copyright 1996-2016 The NASM Authors - All Rights Reserved
+ *
+ *   Copyright 1996-2020 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -14,7 +14,7 @@
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
- *     
+ *
  *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
@@ -31,13 +31,15 @@
  *
  * ----------------------------------------------------------------------- */
 
-/* 
+/*
  * listing.h   header file for listing.c
  */
 
 #ifndef NASM_LISTING_H
 #define NASM_LISTING_H
 
+#include "nasm.h"
+
 /*
  * List-file generators should look like this:
  */
@@ -70,23 +72,27 @@ struct lfmt {
      * `int' parameter is LIST_READ or LIST_MACRO depending on
      * whether the line came directly from an input file or is the
      * result of a multi-line macro expansion.
+     *
+     * If a line number is provided, print it; if the line number is
+     * -1 then use the same line number as the previous call.
      */
-    void (*line)(int type, char *line);
+    void (*line)(int type, int32_t lineno, const char *line);
 
     /*
-     * Called to change one of the various levelled mechanisms in
-     * the listing generator. LIST_INCLUDE and LIST_MACRO can be
-     * used to increase the nesting level of include files and
-     * macro expansions; LIST_TIMES and LIST_INCBIN switch on the
-     * two binary-output-suppression mechanisms for large-scale
-     * pseudo-instructions.
+     * Called to change one of the various levelled mechanisms in the
+     * listing generator. LIST_INCLUDE and LIST_MACRO can be used to
+     * increase the nesting level of include files and macro
+     * expansions; LIST_TIMES and LIST_INCBIN switch on the two
+     * binary-output-suppression mechanisms for large-scale
+     * pseudo-instructions; the size argument prints the size or
+     * repetiiton count.
      *
      * LIST_MACRO_NOLIST is synonymous with LIST_MACRO except that
      * it indicates the beginning of the expansion of a `nolist'
      * macro, so anything under that level won't be expanded unless
      * it includes another file.
      */
-    void (*uplevel)(int type);
+    void (*uplevel)(int type, int64_t size);
 
     /*
      * Reverse the effects of uplevel.
@@ -96,7 +102,7 @@ struct lfmt {
     /*
      * Called on a warning or error, with the error message.
      */
-    void printf_func(2, 3) (*error)(int severity, const char *fmt, ...);
+    void printf_func_ptr(2, 3) (*error)(errflags severity, const char *fmt, ...);
 
     /*
      * Update the current offset.  Used to give the listing generator
@@ -110,4 +116,71 @@ struct lfmt {
 extern const struct lfmt *lfmt;
 extern bool user_nolist;
 
+/*
+ * list_options are the requested options; active_list_options gets
+ * set when a pass starts.
+ *
+ * These are simple bitmasks of ASCII-64 mapping directly to option
+ * letters.
+ */
+extern uint64_t list_options, active_list_options;
+
+/*
+ * This maps the characters a-z, A-Z and 0-9 onto a 64-bit bitmask
+ * (with two bits left over for future use! This isn't particularly
+ * efficient code, but just about every instance of it should be
+ * fed a constant, so the entire function can be precomputed at
+ * compile time. The only cases where the full computation is needed
+ * is when parsing the -L option or %pragma list options, neither of
+ * which is in any way performance critical.
+ *
+ * The character + represents ALL listing options except -Lw (flush
+ * after every line.)
+ *
+ * This returns 0 for invalid values, so that no bit is accessed
+ * for unsupported characters.
+ */
+static inline const_func uint64_t list_option_mask_val(unsigned char x)
+{
+    if (x >= 'a') {
+        if (x > 'z')
+            return 0;
+        x = x - 'a';
+    } else if (x >= 'A') {
+        if (x > 'Z')
+            return 0;
+        x = x - 'A' + 26;
+    } else if (x >= '0') {
+        if (x > '9')
+            return 0;
+        x = x - '0' + 26*2;
+    } else {
+        return 0;
+    }
+
+    return UINT64_C(1) << x;
+}
+
+static inline const_func uint64_t list_option_mask(unsigned char x)
+{
+    if (x == '+')
+        return ~list_option_mask_val('w');
+    else
+        return list_option_mask_val(x);
+}
+
+static inline pure_func bool list_option(unsigned char x)
+{
+    return unlikely(active_list_options & list_option_mask(x));
+}
+
+/* We can't test this using active_list_options for obvious reasons... */
+static inline pure_func bool list_on_every_pass(void)
+{
+    return unlikely(list_options & list_option_mask('p'));
+}
+
+/* Pragma handler */
+enum directive_result list_pragma(const struct pragma *);
+
 #endif
