@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2017 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2020 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -38,9 +38,6 @@
 #include "version.h"
 #include "compiler.h"
 
-#include <stdio.h>
-#include <stddef.h>
-#include <stdlib.h>
 
 #include "nasm.h"
 #include "nasmlib.h"
@@ -62,7 +59,7 @@ static void cv8_output(int type, void *param);
 static void cv8_cleanup(void);
 
 const struct dfmt df_cv8 = {
-    "Codeview 8",               /* .fullname */
+    "Codeview 8+",              /* .fullname */
     "cv8",                      /* .shortname */
     cv8_init,                   /* .init */
     cv8_linenum,                /* .linenum */
@@ -177,7 +174,6 @@ static void cv8_init(void)
 
     cv8_state.source_files = NULL;
     cv8_state.source_files_tail = &cv8_state.source_files;
-    hash_init(&cv8_state.file_hash, HASH_MEDIUM);
 
     cv8_state.num_files = 0;
     cv8_state.total_filename_len = 0;
@@ -309,21 +305,21 @@ static void build_type_table(struct coff_Section *const sect);
 static void cv8_cleanup(void)
 {
     struct cv8_symbol *sym;
-    struct source_file *file;
+    struct source_file *file, *ftmp;
 
     struct coff_Section *symbol_sect = coff_sects[cv8_state.symbol_sect];
     struct coff_Section *type_sect = coff_sects[cv8_state.type_sect];
 
-    cv8_state.outfile.name = nasm_realpath(coff_outfile);
+    cv8_state.outfile.name = nasm_realpath(outname);
     cv8_state.outfile.namebytes = strlen(cv8_state.outfile.name) + 1;
 
     build_symbol_table(symbol_sect);
     build_type_table(type_sect);
 
-    list_for_each(file, cv8_state.source_files) {
+    list_for_each_safe(file, ftmp, cv8_state.source_files) {
         nasm_free(file->fullname);
         saa_free(file->lines);
-        free(file);
+        nasm_free(file);
     }
     hash_free(&cv8_state.file_hash);
 
@@ -369,9 +365,9 @@ done_0:
     fclose(f);
 done:
     if (!success) {
-        nasm_error(ERR_NONFATAL, "unable to hash file %s. "
-                 "Debug information may be unavailable.\n",
-                 filename);
+        nasm_nonfatal("unable to hash file %s. "
+                      "Debug information may be unavailable.",
+                      filename);
     }
     return;
 }
@@ -402,8 +398,7 @@ static struct source_file *register_file(const char *filename)
 
         fullpath = nasm_realpath(filename);
 
-        file = nasm_zalloc(sizeof(*file));
-
+        nasm_new(file);
         file->filename = filename;
         file->fullname = fullpath;
         file->fullnamelen = strlen(fullpath);
@@ -480,7 +475,7 @@ static void register_reloc(struct coff_Section *const sect,
                 return;
         }
     }
-    nasm_panic(0, "codeview: relocation for unregistered symbol: %s", sym);
+    nasm_panic("codeview: relocation for unregistered symbol: %s", sym);
 }
 
 static inline void section_write32(struct coff_Section *sect, uint32_t val)
@@ -613,9 +608,9 @@ static void write_linenumber_table(struct coff_Section *const sect)
     }
 }
 
-static uint16_t write_symbolinfo_obj(struct coff_Section *sect)
+static uint32_t write_symbolinfo_obj(struct coff_Section *sect)
 {
-    uint16_t obj_len;
+    uint32_t obj_len;
 
     obj_len = 2 + 4 + cv8_state.outfile.namebytes;
 
@@ -627,31 +622,43 @@ static uint16_t write_symbolinfo_obj(struct coff_Section *sect)
     return obj_len;
 }
 
-static uint16_t write_symbolinfo_properties(struct coff_Section *sect,
+static uint32_t write_symbolinfo_properties(struct coff_Section *sect,
         const char *const creator_str)
 {
     /* https://github.com/Microsoft/microsoft-pdb/blob/1d60e041/include/cvinfo.h#L3313 */
-    uint16_t creator_len;
+    uint32_t creator_len;
 
     creator_len = 2 + 4 + 2 + 3*2 + 3*2 + strlen(creator_str)+1 + 2;
 
+    /*
+     * We used to use a language ID of 3 for "MASM", since it's closest of the
+     * options available; however, BinScope from WACK (the Windows Application
+     * Certification Kit) tests for specific minimum MASM versions and trying to
+     * match an increasing sequence of random MASM version/build numbers seems
+     * like a fool's errand.
+     *
+     * Instead, use a different language ID (NASM is, after all, not MASM
+     * syntax) and just write the actual NASM version number. BinScope appears
+     * to be happy with that.
+     */
+
     section_write16(sect, creator_len);
     section_write16(sect, 0x1116);
-    section_write32(sect, 3); /* language/flags */
+    section_write32(sect, 'N'); /* language: 'N' (0x4e) for "NASM"; flags are 0 */
     if (win64)
         section_write16(sect, 0x00D0); /* machine */
     else if (win32)
         section_write16(sect, 0x0006); /* machine */
     else
-        nasm_assert(!"neither win32 nor win64 are set!");
+        nasm_panic("neither win32 nor win64 are set!");
     section_write16(sect, 0); /* verFEMajor */
     section_write16(sect, 0); /* verFEMinor */
     section_write16(sect, 0); /* verFEBuild */
 
     /* BinScope/WACK insist on version >= 8.0.50727 */
-    section_write16(sect, 8); /* verMajor */
-    section_write16(sect, 0); /* verMinor */
-    section_write16(sect, 50727); /* verBuild */
+    section_write16(sect, NASM_MAJOR_VER); /* verMajor */
+    section_write16(sect, NASM_MINOR_VER); /* verMinor */
+    section_write16(sect, NASM_SUBMINOR_VER*100 + NASM_PATCHLEVEL_VER); /* verBuild */
 
     section_wbytes(sect, creator_str, strlen(creator_str)+1); /* verSt */
     /*
@@ -663,9 +670,9 @@ static uint16_t write_symbolinfo_properties(struct coff_Section *sect,
     return creator_len;
 }
 
-static uint16_t write_symbolinfo_symbols(struct coff_Section *sect)
+static uint32_t write_symbolinfo_symbols(struct coff_Section *sect)
 {
-    uint16_t len = 0, field_len;
+    uint32_t len = 0, field_len;
     uint32_t field_base;
     struct cv8_symbol *sym;
 
@@ -700,7 +707,7 @@ static uint16_t write_symbolinfo_symbols(struct coff_Section *sect)
             section_write8(sect, 0); /* FLAG */
             break;
         default:
-            nasm_assert(!"unknown symbol type");
+            nasm_panic("unknown symbol type");
         }
 
         section_wbytes(sect, sym->name, strlen(sym->name) + 1);
@@ -719,7 +726,7 @@ static uint16_t write_symbolinfo_symbols(struct coff_Section *sect)
 static void write_symbolinfo_table(struct coff_Section *const sect)
 {
     static const char creator_str[] = "The Netwide Assembler " NASM_VER;
-    uint16_t obj_length, creator_length, sym_length;
+    uint32_t obj_length, creator_length, sym_length;
     uint32_t field_length = 0, out_len;
 
     nasm_assert(cv8_state.outfile.namebytes);
@@ -783,32 +790,28 @@ static void build_symbol_table(struct coff_Section *const sect)
 
 static void build_type_table(struct coff_Section *const sect)
 {
-    uint16_t field_len;
-    struct cv8_symbol *sym;
+    uint32_t field_len;
+    uint32_t typeindex = 0x1000;
+    uint32_t idx_arglist;
 
     section_write32(sect, 0x00000004);
 
-    saa_rewind(cv8_state.symbols);
-    while ((sym = saa_rstruct(cv8_state.symbols))) {
-        if (sym->type != SYMTYPE_PROC)
-            continue;
+    /* empty argument list type */
+    field_len = 2 + 4;
+    section_write16(sect, field_len);
+    section_write16(sect, 0x1201); /* ARGLIST */
+    section_write32(sect, 0); /* num params */
+    idx_arglist = typeindex++;
 
-        /* proc leaf */
+    /* procedure type: void proc(void) */
+    field_len = 2 + 4 + 1 + 1 + 2 + 4;
+    section_write16(sect, field_len);
+    section_write16(sect, 0x1008); /* PROC type */
 
-        field_len = 2 + 4 + 4 + 4 + 2;
-        section_write16(sect, field_len);
-        section_write16(sect, 0x1008); /* PROC type */
-
-        section_write32(sect, 0x00000003); /* return type */
-        section_write32(sect, 0); /* calling convention (default) */
-        section_write32(sect, sym->typeindex);
-        section_write16(sect, 0); /* # params */
-
-        /* arglist */
-
-        field_len = 2 + 4;
-        section_write16(sect, field_len);
-        section_write16(sect, 0x1201); /* ARGLIST */
-        section_write32(sect, 0); /*num params */
-    }
+    section_write32(sect, 0x00000003); /* return type VOID */
+    section_write8(sect, 0);  /* calling convention (default) */
+    section_write8(sect, 0);  /* function attributes */
+    section_write16(sect, 0); /* # params */
+    section_write32(sect, idx_arglist); /* argument list type */
+    /* idx_voidfunc = typeindex++; */
 }
