@@ -38,10 +38,7 @@
 
 #include "compiler.h"
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <ctype.h>
+#include "nctype.h"
 
 #include "nasm.h"
 #include "nasmlib.h"
@@ -159,11 +156,8 @@ static void as86_cleanup(void)
     saa_free(strs);
 }
 
-static int32_t as86_section_names(char *name, int pass, int *bits)
+static int32_t as86_section_names(char *name, int *bits)
 {
-
-    (void)pass;
-
     /*
      * Default is 16 bits.
      */
@@ -200,13 +194,13 @@ static void as86_deflabel(char *name, int32_t segment, int64_t offset,
     struct Symbol *sym;
 
     if (special)
-        nasm_error(ERR_NONFATAL, "as86 format does not support any"
-              " special symbol types");
+        nasm_nonfatal("as86 format does not support any"
+                      " special symbol types");
 
 
     if (name[0] == '.' && name[1] == '.' && name[2] != '@') {
 	if (strcmp(name, "..start")) {
-	    nasm_error(ERR_NONFATAL, "unrecognised special symbol `%s'", name);
+	    nasm_nonfatal("unrecognised special symbol `%s'", name);
 	    return;
 	} else {
 	    is_start = true;
@@ -294,17 +288,7 @@ static void as86_out(int32_t segto, const void *data,
 
     if (wrt != NO_SEG) {
         wrt = NO_SEG;           /* continue to do _something_ */
-        nasm_error(ERR_NONFATAL, "WRT not supported by as86 output format");
-    }
-
-    /*
-     * handle absolute-assembly (structure definitions)
-     */
-    if (segto == NO_SEG) {
-        if (type != OUT_RESERVE)
-            nasm_error(ERR_NONFATAL, "attempt to assemble code in [ABSOLUTE]"
-                  " space");
-        return;
+        nasm_nonfatal("WRT not supported by as86 output format");
     }
 
     if (segto == stext.index)
@@ -314,14 +298,14 @@ static void as86_out(int32_t segto, const void *data,
     else if (segto == bssindex)
         s = NULL;
     else {
-        nasm_error(ERR_WARNING, "attempt to assemble code in"
-              " segment %d: defaulting to `.text'", segto);
+        nasm_warn(WARN_OTHER, "attempt to assemble code in"
+                  " segment %d: defaulting to `.text'", segto);
         s = &stext;
     }
 
     if (!s && type != OUT_RESERVE) {
-        nasm_error(ERR_WARNING, "attempt to initialize memory in the"
-              " BSS section: ignored");
+        nasm_warn(WARN_OTHER, "attempt to initialize memory in the"
+                  " BSS section: ignored");
 	bsslen += realsize(type, size);
         return;
     }
@@ -330,24 +314,22 @@ static void as86_out(int32_t segto, const void *data,
 
     if (type == OUT_RESERVE) {
         if (s) {
-            nasm_error(ERR_WARNING, "uninitialized space declared in"
-                  " %s section: zeroing",
-                  (segto == stext.index ? "code" : "data"));
+            nasm_warn(WARN_ZEROING, "uninitialized space declared in"
+                      " %s section: zeroing",
+                      (segto == stext.index ? "code" : "data"));
             as86_sect_write(s, NULL, size);
             as86_add_piece(s, 0, 0L, 0L, size, 0);
         } else
             bsslen += size;
     } else if (type == OUT_RAWDATA) {
-        if (segment != NO_SEG)
-            nasm_panic(0, "OUT_RAWDATA with other than NO_SEG");
         as86_sect_write(s, data, size);
         as86_add_piece(s, 0, 0L, 0L, size, 0);
     } else if (type == OUT_ADDRESS) {
         int asize = abs((int)size);
         if (segment != NO_SEG) {
             if (segment % 2) {
-                nasm_error(ERR_NONFATAL, "as86 format does not support"
-                      " segment base references");
+                nasm_nonfatal("as86 format does not support"
+                              " segment base references");
             } else {
                 offset = *(int64_t *)data;
                 as86_add_piece(s, 1, offset, segment, asize, 0);
@@ -359,12 +341,10 @@ static void as86_out(int32_t segto, const void *data,
             as86_add_piece(s, 0, 0L, 0L, asize, 0);
         }
     } else if (type == OUT_REL2ADR) {
-        if (segment == segto)
-            nasm_panic(0, "intra-segment OUT_REL2ADR");
         if (segment != NO_SEG) {
             if (segment % 2) {
-                nasm_error(ERR_NONFATAL, "as86 format does not support"
-                      " segment base references");
+                nasm_nonfatal("as86 format does not support"
+                              " segment base references");
             } else {
                 offset = *(int64_t *)data;
                 as86_add_piece(s, 1, offset - size + 2, segment, 2L,
@@ -372,12 +352,10 @@ static void as86_out(int32_t segto, const void *data,
             }
         }
     } else if (type == OUT_REL4ADR) {
-        if (segment == segto)
-            nasm_panic(0, "intra-segment OUT_REL4ADR");
         if (segment != NO_SEG) {
             if (segment % 2) {
-                nasm_error(ERR_NONFATAL, "as86 format does not support"
-                      " segment base references");
+                nasm_nonfatal("as86 format does not support"
+                              " segment base references");
             } else {
                 offset = *(int64_t *)data;
                 as86_add_piece(s, 1, offset - size + 4, segment, 4L,
@@ -514,7 +492,7 @@ static void as86_set_rsize(int size)
             fputc(0x03, ofile);
             break;
         default:
-            nasm_panic(0, "bizarre relocation size %d", size);
+            nasm_panic("bizarre relocation size %d", size);
 	    break;
         }
     }
@@ -605,15 +583,10 @@ static void as86_sect_write(struct Section *sect,
     sect->datalen += len;
 }
 
-static int32_t as86_segbase(int32_t segment)
-{
-    return segment;
-}
-
 extern macros_t as86_stdmac[];
 
 const struct ofmt of_as86 = {
-    "Linux as86 (bin86 version 0.3) object files",
+    "as86 (bin86/dev86 toolchain)",
     "as86",
     ".o",
     0,
@@ -629,7 +602,7 @@ const struct ofmt of_as86 = {
     as86_section_names,
     NULL,
     null_sectalign,
-    as86_segbase,
+    null_segbase,
     null_directive,
     as86_cleanup,
     NULL                        /* pragma list */
