@@ -38,10 +38,7 @@
 
 #include "compiler.h"
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <ctype.h>
+#include "nctype.h"
 #include <errno.h>
 
 #include "nasm.h"
@@ -75,8 +72,8 @@ static void dbg_init(void)
 
 static void dbg_reset(void)
 {
-    fprintf(ofile, "*** pass reset: pass0 = %d, passn = %"PRId64"\n",
-            pass0, passn);
+    fprintf(ofile, "*** pass reset: pass = %"PRId64" (%s)\n",
+            pass_count(), pass_type_name());
 }
 
 static void dbg_cleanup(void)
@@ -90,8 +87,7 @@ static void dbg_cleanup(void)
     }
 }
 
-static int32_t dbg_add_section(char *name, int pass, int *bits,
-                                     const char *whatwecallit)
+static int32_t dbg_add_section(char *name, int *bits, const char *whatwecallit)
 {
     int seg;
 
@@ -121,8 +117,8 @@ static int32_t dbg_add_section(char *name, int pass, int *bits,
             s->number = seg = seg_alloc();
             s->next = dbgsect;
             dbgsect = s;
-            fprintf(ofile, "%s %s (%s) pass %d: returning %d\n",
-                    whatwecallit, name, tail, pass, seg);
+            fprintf(ofile, "%s %s (%s) pass %"PRId64" (%s) : returning %d\n",
+                    whatwecallit, name, tail, pass_count(), pass_type_name(), seg);
 
             if (section_labels)
                 backend_label(s->name, s->number + 1, 0);
@@ -131,9 +127,9 @@ static int32_t dbg_add_section(char *name, int pass, int *bits,
     return seg;
 }
 
-static int32_t dbg_section_names(char *name, int pass, int *bits)
+static int32_t dbg_section_names(char *name, int *bits)
 {
-    return dbg_add_section(name, pass, bits, "section_names");
+    return dbg_add_section(name, bits, "section_names");
 }
 
 static int32_t dbg_herelabel(const char *name, enum label_type type,
@@ -322,13 +318,8 @@ static void dbg_sectalign(int32_t seg, unsigned int value)
             seg, value);
 }
 
-static int32_t dbg_segbase(int32_t segment)
-{
-    return segment;
-}
-
 static enum directive_result
-dbg_directive(enum directive directive, char *value, int pass)
+dbg_directive(enum directive directive, char *value)
 {
     switch (directive) {
         /*
@@ -339,7 +330,7 @@ dbg_directive(enum directive directive, char *value, int pass)
     case D_GROUP:
     {
         int dummy;
-        dbg_add_section(value, pass, &dummy, "directive:group");
+        dbg_add_section(value, &dummy, "directive:group");
         break;
     }
 
@@ -347,8 +338,8 @@ dbg_directive(enum directive directive, char *value, int pass)
         break;
     }
 
-    fprintf(ofile, "directive [%s] value [%s] (pass %d)\n",
-            directive_dname(directive), value, pass);
+    fprintf(ofile, "directive [%s] value [%s] pass %"PRId64" (%s)\n",
+            directive_dname(directive), value, pass_count(), pass_type_name());
     return DIRR_OK;
 }
 
@@ -380,7 +371,7 @@ dbg_pragma(const struct pragma *pragma)
                 errno = 0;
                 arg = strtoul(pragma->tail, &ep, 0);
                 if (errno || *nasm_skip_spaces(ep)) {
-                    nasm_error(ERR_WARNING | WARN_BAD_PRAGMA | ERR_PASS2,
+                    nasm_warn(WARN_PRAGMA_BAD | ERR_PASS2,
                                "invalid %%pragma dbg maxdump argument");
                     return DIRR_ERROR;
                 } else {
@@ -484,7 +475,7 @@ const struct ofmt of_dbg = {
     dbg_section_names,
     dbg_herelabel,
     dbg_sectalign,
-    dbg_segbase,
+    null_segbase,
     dbg_directive,
     dbg_cleanup,
     dbg_pragma_list
