@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2017 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2018 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -38,22 +38,22 @@
 
 #include "compiler.h"
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <ctype.h>
+#include "nctype.h"
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "ilog2.h"
 #include "labels.h"
 #include "error.h"
 #include "saa.h"
 #include "raa.h"
 #include "rbtree.h"
+#include "hashtbl.h"
 #include "outform.h"
 #include "outlib.h"
 #include "ver.h"
 #include "dwarf.h"
+#include "macho.h"
 
 #if defined(OF_MACHO) || defined(OF_MACHO64)
 
@@ -70,45 +70,8 @@
 #define MACHO_SECTCMD64_SIZE		80
 #define MACHO_NLIST64_SIZE		16
 
-/* Mach-O file header values */
-#define	MH_MAGIC		0xfeedface
-#define	MH_MAGIC_64		0xfeedfacf
-#define CPU_TYPE_I386		7		/* x86 platform */
-#define CPU_TYPE_X86_64		0x01000007	/* x86-64 platform */
-#define	CPU_SUBTYPE_I386_ALL	3		/* all-x86 compatible */
-#define	MH_OBJECT		0x1		/* object file */
-
-/* Mach-O header flags */
-#define MH_SUBSECTIONS_VIA_SYMBOLS 0x2000
-
-/* Mach-O load commands */
-#define LC_SEGMENT		0x1		/* 32-bit segment load cmd */
-#define LC_SEGMENT_64		0x19		/* 64-bit segment load cmd */
-#define LC_SYMTAB		0x2		/* symbol table load command */
-
 /* Mach-O relocations numbers */
 
-/* Generic relocs, used by i386 Mach-O */
-#define GENERIC_RELOC_VANILLA   0               /* Generic relocation */
-#define GENERIC_RELOC_TLV	5		/* Thread local */
-
-#define X86_64_RELOC_UNSIGNED   0               /* Absolute address */
-#define X86_64_RELOC_SIGNED     1               /* Signed 32-bit disp */
-#define X86_64_RELOC_BRANCH     2		/* CALL/JMP with 32-bit disp */
-#define X86_64_RELOC_GOT_LOAD   3		/* MOVQ of GOT entry */
-#define X86_64_RELOC_GOT        4		/* Different GOT entry */
-#define X86_64_RELOC_SUBTRACTOR 5		/* Subtracting two symbols */
-#define X86_64_RELOC_SIGNED_1   6		/* SIGNED with -1 addend */
-#define X86_64_RELOC_SIGNED_2   7		/* SIGNED with -2 addend */
-#define X86_64_RELOC_SIGNED_4   8		/* SIGNED with -4 addend */
-#define X86_64_RELOC_TLV        9		/* Thread local */
-
-/* Mach-O VM permission constants */
-#define	VM_PROT_NONE	(0x00)
-#define VM_PROT_READ	(0x01)
-#define VM_PROT_WRITE	(0x02)
-#define VM_PROT_EXECUTE	(0x04)
-
 #define VM_PROT_DEFAULT	(VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE)
 #define VM_PROT_ALL	(VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE)
 
@@ -152,16 +115,18 @@ struct section {
     /* nasm internal data */
     struct section *next;
     struct SAA *data;
-    int32_t index;
+    int32_t index;		/* Main section index */
+    int32_t subsection;		/* Current subsection index */
     int32_t fileindex;
     struct reloc *relocs;
     struct rbtree *syms[2]; /* All/global symbols symbols in section */
     int align;
-    bool by_name;		/* This section was specified by full MachO name */
+    bool by_name;	    /* This section was specified by full MachO name */
+    char namestr[34];	    /* segment,section as a C string */
 
     /* data that goes into the file */
-    char sectname[16];          /* what this section is called */
-    char segname[16];           /* segment this section will be in */
+    char sectname[16];     /* what this section is called */
+    char segname[16];      /* segment this section will be in */
     uint64_t addr;         /* in-memory address (subject to alignment) */
     uint64_t size;         /* in-memory and -file size  */
     uint64_t offset;	   /* in-file offset */
@@ -171,25 +136,6 @@ struct section {
     uint32_t extreloc;     /* external relocations */
 };
 
-#define SECTION_TYPE	0x000000ff      /* section type mask */
-
-#define	S_REGULAR		(0x0)   /* standard section */
-#define	S_ZEROFILL		(0x1)   /* zerofill, in-memory only */
-
-#define SECTION_ATTRIBUTES_SYS   0x00ffff00     /* system setable attributes */
-#define S_ATTR_SOME_INSTRUCTIONS 0x00000400     /* section contains some
-						   machine instructions */
-#define S_ATTR_EXT_RELOC         0x00000200     /* section has external relocation entries */
-#define S_ATTR_LOC_RELOC         0x00000100     /* section has local relocation entries */
-#define S_ATTR_DEBUG		 0x02000000
-#define S_ATTR_SELF_MODIFYING_CODE 0x04000000
-#define S_ATTR_LIVE_SUPPORT	 0x08000000
-#define S_ATTR_NO_DEAD_STRIP	 0x10000000     /* no dead stripping */
-#define S_ATTR_STRIP_STATIC_SYMS 0x20000000
-#define S_ATTR_NO_TOC            0x40000000
-#define S_ATTR_PURE_INSTRUCTIONS 0x80000000	/* section uses pure machine instructions */
-#define S_ATTR_DEBUG 0x02000000 /* debug section */
-
 #define S_NASM_TYPE_MASK	 0x800004ff	/* we consider these bits "section type" */
 
 /* fake section for absolute symbols, *not* part of the section linked list */
@@ -210,10 +156,6 @@ struct reloc {
 	type:4;                 /* reloc type */
 };
 
-#define	R_ABS		0       /* absolute relocation */
-#define R_SCATTERED	0x80000000      /* reloc entry is scattered if
-					** highest bit == 1 */
-
 struct symbol {
     /* nasm internal data */
     struct rbtree symv[2];	/* All/global symbol rbtrees; "key" contains the
@@ -230,22 +172,8 @@ struct symbol {
     uint16_t desc;		/* for stab debugging, 0 for us */
 };
 
-/* symbol type bits */
-#define	N_EXT	0x01            /* global or external symbol */
-
-#define	N_UNDF	0x0             /* undefined symbol | n_sect == */
-#define	N_ABS	0x2             /* absolute symbol  |  NO_SECT */
-#define	N_SECT	0xe             /* defined symbol, n_sect holds
-				** section number */
-
-#define	N_TYPE	0x0e            /* type bit mask */
-
 #define DEFAULT_SECTION_ALIGNMENT 0 /* byte (i.e. no) alignment */
 
-/* special section number values */
-#define	NO_SECT		0       /* no section, invalid */
-#define MAX_SECT	255     /* maximum number of sections */
-
 static struct section *sects, **sectstail, **sectstab;
 static struct symbol *syms, **symstail;
 static uint32_t nsyms;
@@ -307,27 +235,58 @@ static uint64_t rel_padcnt = 0;
 #define alignptr(x) \
     ALIGN(x, fmt.ptrsize)	/* align x to output format width */
 
-static struct section *get_section_by_name(const char *segname,
-                                           const char *sectname)
-{
-    struct section *s;
+static struct hash_table section_by_name;
+static struct RAA *section_by_index;
 
-    for (s = sects; s != NULL; s = s->next)
-        if (!xstrncmp(s->segname, segname) && !xstrncmp(s->sectname, sectname))
-            break;
+static struct section * never_null
+find_or_add_section(const char *segname, const char *sectname)
+{
+    struct hash_insert hi;
+    void **sp;
+    struct section *s;
+    char sect[34];
+
+    snprintf(sect, sizeof sect, "%-16s,%-16s", segname, sectname);
+
+    sp = hash_find(&section_by_name, sect, &hi);
+    if (sp)
+	return (struct section *)(*sp);
+
+    s = nasm_zalloc(sizeof *s);
+    xstrncpy(s->segname, segname);
+    xstrncpy(s->sectname, sectname);
+    xstrncpy(s->namestr, sect);
+    hash_add(&hi, s->namestr, s);
+
+    s->index = s->subsection = seg_alloc();
+    section_by_index = raa_write_ptr(section_by_index, s->index >> 1, s);
 
     return s;
 }
 
-static struct section *get_section_by_index(const int32_t index)
+static inline bool is_new_section(const struct section *s)
 {
-    struct section *s;
+    return !s->data;
+}
 
-    for (s = sects; s != NULL; s = s->next)
-        if (index == s->index)
-            break;
+static struct section *get_section_by_name(const char *segname,
+                                           const char *sectname)
+{
+    char sect[34];
+    void **sp;
 
-    return s;
+    snprintf(sect, sizeof sect, "%-16s,%-16s", segname, sectname);
+
+    sp = hash_find(&section_by_name, sect, NULL);
+    return sp ? (struct section *)(*sp) : NULL;
+}
+
+static struct section *get_section_by_index(int32_t index)
+{
+    if (index < 0 || index >= SEG_ABS || (index & 1))
+	return NULL;
+
+    return raa_read_ptr(section_by_index, index >> 1);
 }
 
 struct dir_list {
@@ -383,6 +342,7 @@ static int32_t macho_gotpcrel_sect;
 
 static void macho_init(void)
 {
+    module_name = inname;
     sects = NULL;
     sectstail = &sects;
 
@@ -399,14 +359,15 @@ static void macho_init(void)
     extsyms = raa_init();
     strs = saa_init(1L);
 
+    section_by_index = raa_init();
+
     /* string table starts with a zero byte so index 0 is an empty string */
     saa_wbytes(strs, zero_buffer, 1);
     strslen = 1;
 
     /* add special symbol for TLVP */
     macho_tlvp_sect = seg_alloc() + 1;
-    define_label("..tlvp", macho_tlvp_sect, 0L, NULL, false, false);
-
+    backend_label("..tlvp", macho_tlvp_sect, 0L);
 }
 
 static void sect_write(struct section *sect,
@@ -456,7 +417,7 @@ static int64_t add_reloc(struct section *sect, int32_t section,
     r = nasm_malloc(sizeof(struct reloc));
     r->addr = sect->size & ~R_SCATTERED;
     r->ext = 1;
-    adjust = bytes;
+    adjust = 0;
 
     /* match byte count 1, 2, 4, 8 to length codes 0, 1, 2, 3 respectively */
     r->length = ilog2_32(bytes);
@@ -466,9 +427,7 @@ static int64_t add_reloc(struct section *sect, int32_t section,
     r->pcrel = 0;
     r->snum = R_ABS;
 
-    s = NULL;
-    if (section != NO_SEG)
-	s = get_section_by_index(section);
+    s = get_section_by_index(section);
     fi = s ? s->fileindex : NO_SECT;
 
     /* absolute relocation */
@@ -477,7 +436,6 @@ static int64_t add_reloc(struct section *sect, int32_t section,
 	if (section == NO_SEG) {
 	    /* absolute (can this even happen?) */
 	    r->ext = 0;
-	    r->snum = NO_SECT;
 	} else if (fi == NO_SECT) {
 	    /* external */
 	    r->snum = raa_read(extsyms, section);
@@ -485,7 +443,6 @@ static int64_t add_reloc(struct section *sect, int32_t section,
 	    /* local */
 	    r->ext = 0;
 	    r->snum = fi;
-	    adjust = -sect->size;
 	}
 	break;
 
@@ -494,40 +451,27 @@ static int64_t add_reloc(struct section *sect, int32_t section,
 	r->type = fmt.reloc_rel;
 	r->pcrel = 1;
 	if (section == NO_SEG) {
-	    /* absolute - seems to produce garbage no matter what */
-	    nasm_error(ERR_NONFATAL, "Mach-O does not support relative "
-		       "references to absolute addresses");
-	    goto bail;
-#if 0
-	    /* This "seems" to be how it ought to work... */
-
-	    struct symbol *sym = macho_find_sym(&absolute_sect, offset,
-						false, false);
-	    if (!sym)
-		goto bail;
-
-	    sect->extreloc = 1;
-	    r->snum = NO_SECT;
-	    adjust = -sect->size;
-#endif
+	    /* may optionally be converted below by fmt.forcesym */
+	    r->ext = 0;
 	} else if (fi == NO_SECT) {
 	    /* external */
 	    sect->extreloc = 1;
 	    r->snum = raa_read(extsyms, section);
 	    if (reltype == RL_BRANCH)
 		r->type = X86_64_RELOC_BRANCH;
-	    else if (r->type == GENERIC_RELOC_VANILLA)
-		adjust = -sect->size;
 	} else {
 	    /* local */
 	    r->ext = 0;
 	    r->snum = fi;
-	    adjust = -sect->size;
+	    if (reltype == RL_BRANCH)
+		r->type = X86_64_RELOC_BRANCH;
 	}
 	break;
 
-    case RL_SUB:
-	r->pcrel = 0;
+    case RL_SUB: /* obsolete */
+	nasm_warn(WARN_OTHER, "relcation with subtraction"
+		   "becomes to be obsolete");
+	r->ext = 0;
 	r->type = X86_64_RELOC_SUBTRACTOR;
 	break;
 
@@ -544,34 +488,46 @@ static int64_t add_reloc(struct section *sect, int32_t section,
 	goto needsym;
 
     needsym:
-	r->pcrel = 1;
+	r->pcrel = (fmt.ptrsize == 8 ? 1 : 0);
 	if (section == NO_SEG) {
 	    nasm_error(ERR_NONFATAL, "Unsupported use of use of WRT");
+	    goto bail;
 	} else if (fi == NO_SECT) {
 	    /* external */
 	    r->snum = raa_read(extsyms, section);
 	} else {
-	    /* internal - does it really need to be global? */
-	    struct symbol *sym = macho_find_sym(s, offset, true,
-						reltype != RL_TLV);
-	    if (!sym)
+	    /* internal - GOTPCREL doesn't need to be in global */
+	    struct symbol *sym = macho_find_sym(s, offset,
+						false, /* reltype != RL_TLV */
+						true);
+	    if (!sym) {
+		nasm_error(ERR_NONFATAL, "Symbol for WRT not found");
 		goto bail;
+	    }
 
+	    adjust -= sym->symv[0].key;
 	    r->snum = sym->initial_snum;
 	}
 	break;
     }
 
-    /* For 64-bit Mach-O, force a symbol reference if at all possible */
-    if (!r->ext && r->snum != NO_SECT && fmt.forcesym) {
-	struct symbol *sym = macho_find_sym(s, offset, false, false);
+    /*
+     * For 64-bit Mach-O, force a symbol reference if at all possible
+     * Allow for r->snum == R_ABS by searching absolute_sect
+     */
+    if (!r->ext && fmt.forcesym) {
+	struct symbol *sym = macho_find_sym(s ? s : &absolute_sect,
+					    offset, false, false);
 	if (sym) {
-	    adjust = bytes - sym->symv[0].key;
+	    adjust -= sym->symv[0].key;
 	    r->snum = sym->initial_snum;
 	    r->ext = 1;
 	}
     }
 
+    if (r->pcrel)
+	adjust += ((r->ext && fmt.ptrsize == 8) ? bytes : -(int64_t)sect->size);
+
     /* NeXT as puts relocs in reversed order (address-wise) into the
      ** files, so we do the same, doesn't seem to make much of a
      ** difference either way */
@@ -598,23 +554,15 @@ static void macho_output(int32_t secto, const void *data,
     bool is_bss;
     enum reltype reltype;
 
-    if (secto == NO_SEG) {
-        if (type != OUT_RESERVE)
-            nasm_error(ERR_NONFATAL, "attempt to assemble code in "
-                  "[ABSOLUTE] space");
-        return;
-    }
-
     s = get_section_by_index(secto);
-
-    if (s == NULL) {
-        nasm_error(ERR_WARNING, "attempt to assemble code in"
+    if (!s) {
+        nasm_warn(WARN_OTHER, "attempt to assemble code in"
               " section %d: defaulting to `.text'", secto);
         s = get_section_by_name("__TEXT", "__text");
 
         /* should never happen */
-        if (s == NULL)
-            nasm_panic(0, "text section not found");
+        if (!s)
+            nasm_panic("text section not found");
     }
 
     /* debug code generation only for sections tagged with
@@ -630,8 +578,11 @@ static void macho_output(int32_t secto, const void *data,
     is_bss = (s->flags & SECTION_TYPE) == S_ZEROFILL;
 
     if (is_bss && type != OUT_RESERVE) {
-        nasm_error(ERR_WARNING, "attempt to initialize memory in "
+        nasm_warn(WARN_OTHER, "attempt to initialize memory in "
               "BSS section: ignored");
+        /* FIXME */
+        nasm_warn(WARN_OTHER, "section size may be negative"
+            "with address symbols");
         s->size += realsize(type, size);
         return;
     }
@@ -641,7 +592,7 @@ static void macho_output(int32_t secto, const void *data,
     switch (type) {
     case OUT_RESERVE:
         if (!is_bss) {
-            nasm_error(ERR_WARNING, "uninitialized space declared in"
+            nasm_warn(WARN_ZEROING, "uninitialized space declared in"
 		       " %s,%s section: zeroing", s->segname, s->sectname);
 
             sect_write(s, NULL, size);
@@ -651,9 +602,6 @@ static void macho_output(int32_t secto, const void *data,
         break;
 
     case OUT_RAWDATA:
-        if (section != NO_SEG)
-            nasm_panic(0, "OUT_RAWDATA with other than NO_SEG");
-
         sect_write(s, data, size);
         break;
 
@@ -672,8 +620,11 @@ static void macho_output(int32_t secto, const void *data,
 			       "Mach-O 64-bit format does not support"
 			       " 32-bit absolute addresses");
 		} else {
-		    add_reloc(s, section, addr, RL_ABS, asize);
+		    addr += add_reloc(s, section, addr, RL_ABS, asize);
 		}
+	    } else if (wrt == macho_tlvp_sect && fmt.ptrsize != 8 &&
+		       asize == (int) fmt.ptrsize) {
+		addr += add_reloc(s, section, addr, RL_TLV, asize);
 	    } else {
 		nasm_error(ERR_NONFATAL, "Mach-O format does not support"
 			   " this use of WRT");
@@ -686,8 +637,8 @@ static void macho_output(int32_t secto, const void *data,
         break;
     }
 
+    case OUT_REL1ADR:
     case OUT_REL2ADR:
-	nasm_assert(section != secto);
 
         p = mydata;
 	offset = *(int64_t *)data;
@@ -704,15 +655,16 @@ static void macho_output(int32_t secto, const void *data,
 		       " this use of WRT");
 	    wrt = NO_SEG;	/* we can at least _try_ to continue */
 	} else {
-	    addr += add_reloc(s, section, addr+size, RL_REL, 2);
+	    addr += add_reloc(s, section, addr+size, RL_REL,
+                              type == OUT_REL1ADR ? 1 : 2);
 	}
 
         WRITESHORT(p, addr);
-        sect_write(s, mydata, 2);
+        sect_write(s, mydata, type == OUT_REL1ADR ? 1 : 2);
         break;
 
     case OUT_REL4ADR:
-	nasm_assert(section != secto);
+    case OUT_REL8ADR:
 
         p = mydata;
 	offset = *(int64_t *)data;
@@ -758,7 +710,7 @@ static void macho_output(int32_t secto, const void *data,
 		    reltype = RL_GOTLOAD;
 		}
 	    }
-	} else if (wrt == macho_tlvp_sect) {
+	} else if (wrt == macho_tlvp_sect && fmt.ptrsize == 8) {
 	    reltype = RL_TLV;
 	} else {
 	    nasm_error(ERR_NONFATAL, "Mach-O format does not support"
@@ -766,9 +718,10 @@ static void macho_output(int32_t secto, const void *data,
 	    /* continue with RL_REL */
 	}
 
-	addr += add_reloc(s, section, offset, reltype, 4);
+	addr += add_reloc(s, section, offset, reltype,
+                          type == OUT_REL4ADR ? 4 : 8);
         WRITELONG(p, addr);
-        sect_write(s, mydata, 4);
+        sect_write(s, mydata, type == OUT_REL4ADR ? 4 : 8);
         break;
 
     default:
@@ -777,59 +730,74 @@ static void macho_output(int32_t secto, const void *data,
     }
 }
 
-/* Translation table from traditional Unix section names to Mach-O */
-static const struct sectmap {
-    const char *nasmsect;
-    const char *segname;
-    const char *sectname;
-    const uint32_t flags;
-} sectmap[] = {
-    {".text", "__TEXT", "__text",
-     S_REGULAR|S_ATTR_SOME_INSTRUCTIONS|S_ATTR_PURE_INSTRUCTIONS},
-    {".data", "__DATA", "__data", S_REGULAR},
-    {".rodata", "__DATA", "__const", S_REGULAR},
-    {".bss", "__DATA", "__bss", S_ZEROFILL},
-    {".debug_abbrev", "__DWARF", "__debug_abbrev", S_ATTR_DEBUG},
-    {".debug_info", "__DWARF", "__debug_info", S_ATTR_DEBUG},
-    {".debug_line", "__DWARF", "__debug_line", S_ATTR_DEBUG},
-    {".debug_str", "__DWARF", "__debug_str", S_ATTR_DEBUG},
-    {NULL, NULL, NULL, 0}
-};
-
+#define S_CODE  (S_REGULAR | S_ATTR_SOME_INSTRUCTIONS | S_ATTR_PURE_INSTRUCTIONS)
 #define NO_TYPE S_NASM_TYPE_MASK
 
+/* Translation table from traditional Unix section names to Mach-O */
+static const struct macho_known_section {
+    const char      *nasmsect;
+    const char      *segname;
+    const char      *sectname;
+    const uint32_t  flags;
+} known_sections[] = {
+    { ".text",          "__TEXT",   "__text",           S_CODE          },
+    { ".data",          "__DATA",   "__data",           S_REGULAR       },
+    { ".rodata",        "__DATA",   "__const",          S_REGULAR       },
+    { ".bss",           "__DATA",   "__bss",            S_ZEROFILL      },
+    { ".debug_abbrev",  "__DWARF",  "__debug_abbrev",   S_ATTR_DEBUG    },
+    { ".debug_info",    "__DWARF",  "__debug_info",     S_ATTR_DEBUG    },
+    { ".debug_line",    "__DWARF",  "__debug_line",     S_ATTR_DEBUG    },
+    { ".debug_str",     "__DWARF",  "__debug_str",      S_ATTR_DEBUG    },
+};
+
 /* Section type or attribute directives */
-static const struct sect_attribs {
-    const char *name;
-    uint32_t flags;
+static const struct macho_known_section_attr {
+    const char      *name;
+    uint32_t        flags;
 } sect_attribs[] = {
-    { "data", S_REGULAR },
-    { "code", S_REGULAR|S_ATTR_SOME_INSTRUCTIONS|S_ATTR_PURE_INSTRUCTIONS },
-    { "mixed", S_REGULAR|S_ATTR_SOME_INSTRUCTIONS },
-    { "bss", S_ZEROFILL },
-    { "zerofill", S_ZEROFILL },
-    { "no_dead_strip", NO_TYPE|S_ATTR_NO_DEAD_STRIP },
-    { "live_support", NO_TYPE|S_ATTR_LIVE_SUPPORT },
-    { "strip_static_syms", NO_TYPE|S_ATTR_STRIP_STATIC_SYMS },
-    { "debug", NO_TYPE|S_ATTR_DEBUG },
+    { "data",               S_REGULAR                               },
+    { "code",               S_CODE                                  },
+    { "mixed",              S_REGULAR | S_ATTR_SOME_INSTRUCTIONS    },
+    { "bss",                S_ZEROFILL                              },
+    { "zerofill",           S_ZEROFILL                              },
+    { "no_dead_strip",      NO_TYPE | S_ATTR_NO_DEAD_STRIP          },
+    { "live_support",       NO_TYPE | S_ATTR_LIVE_SUPPORT           },
+    { "strip_static_syms",  NO_TYPE | S_ATTR_STRIP_STATIC_SYMS      },
+    { "debug",              NO_TYPE | S_ATTR_DEBUG                  },
     { NULL, 0 }
 };
 
-static int32_t macho_section(char *name, int pass, int *bits)
+static const struct macho_known_section *
+lookup_known_section(const char *name, bool by_sectname)
 {
+    size_t i;
+
+    if (name && name[0]) {
+            for (i = 0; i < ARRAY_SIZE(known_sections); i++) {
+                const char *p = by_sectname ?
+                    known_sections[i].sectname :
+                    known_sections[i].nasmsect;
+                if (!strcmp(name, p))
+                    return &known_sections[i];
+            }
+    }
+
+    return NULL;
+}
+
+static int32_t macho_section(char *name, int *bits)
+{
+    const struct macho_known_section *known_section;
+    const struct macho_known_section_attr *sa;
     char *sectionAttributes;
-    const struct sectmap *sm;
     struct section *s;
     const char *section, *segment;
     uint32_t flags;
-    const struct sect_attribs *sa;
     char *currentAttribute;
     char *comma;
 
     bool new_seg;
 
-    (void)pass;
-
     /* Default to the appropriate number of bits. */
     if (!name) {
         *bits = fmt.ptrsize << 3;
@@ -865,54 +833,42 @@ static int32_t macho_section(char *name, int pass, int *bits)
 	    nasm_error(ERR_NONFATAL, "section name %s too long\n", section);
 	}
 
-	if (!strcmp(section, "__text")) {
-	    flags = S_REGULAR | S_ATTR_SOME_INSTRUCTIONS |
-		S_ATTR_PURE_INSTRUCTIONS;
-	} else if (!strcmp(section, "__bss")) {
-	    flags = S_ZEROFILL;
-	} else {
-	    flags = S_REGULAR;
-	}
+        known_section = lookup_known_section(section, true);
+        if (known_section)
+            flags = known_section->flags;
+        else
+            flags = S_REGULAR;
     } else {
-	for (sm = sectmap; sm->nasmsect != NULL; ++sm) {
-	    /* make lookup into section name translation table */
-	    if (!strcmp(name, sm->nasmsect)) {
-		segment = sm->segname;
-		section = sm->sectname;
-		flags = sm->flags;
-		goto found;
-	    }
-	}
-	nasm_error(ERR_NONFATAL, "unknown section name\n");
-	return NO_SEG;
+        known_section = lookup_known_section(name, false);
+        if (!known_section) {
+            nasm_error(ERR_NONFATAL, "unknown section name %s\n", name);
+            return NO_SEG;
+        }
+
+        segment = known_section->segname;
+        section = known_section->sectname;
+        flags = known_section->flags;
     }
 
- found:
-    /* try to find section with that name */
-    s = get_section_by_name(segment, section);
+    /* try to find section with that name, or create it */
+    s = find_or_add_section(segment, section);
+    new_seg = is_new_section(s);
 
-    /* create it if it doesn't exist yet */
-    if (!s) {
-	new_seg = true;
-
-	s = *sectstail = nasm_zalloc(sizeof(struct section));
+    /* initialize it if it is a brand new section */
+    if (new_seg) {
+	*sectstail = s;
 	sectstail = &s->next;
 
 	s->data = saa_init(1L);
-	s->index = seg_alloc();
 	s->fileindex = ++seg_nsects;
 	s->align = -1;
 	s->pad = -1;
 	s->offset = -1;
 	s->by_name = false;
 
-	xstrncpy(s->segname, segment);
-	xstrncpy(s->sectname, section);
 	s->size = 0;
 	s->nreloc = 0;
 	s->flags = flags;
-    } else {
-	new_seg = false;
     }
 
     if (comma)
@@ -982,7 +938,38 @@ static int32_t macho_section(char *name, int pass, int *bits)
 	s->flags |= flags & ~S_NASM_TYPE_MASK;
     }
 
-    return s->index;
+    return s->subsection;
+}
+
+static int32_t macho_herelabel(const char *name, enum label_type type,
+			       int32_t section, int32_t *subsection,
+			       bool *copyoffset)
+{
+    struct section *s;
+    int32_t subsec;
+    (void)name;
+
+    if (!(head_flags & MH_SUBSECTIONS_VIA_SYMBOLS))
+	return section;
+
+    /* No subsection only for local labels */
+    if (type == LBL_LOCAL)
+	return section;
+
+    s = get_section_by_index(section);
+    if (!s)
+	return section;
+
+    subsec = *subsection;
+    if (subsec == NO_SEG) {
+	/* Allocate a new subsection index */
+	subsec = *subsection = seg_alloc();
+	section_by_index = raa_write_ptr(section_by_index, subsec >> 1, s);
+    }
+
+    s->subsection = subsec;
+    *copyoffset = true;		/* Maintain previous offset */
+    return subsec;
 }
 
 static void macho_symdef(char *name, int32_t section, int64_t offset,
@@ -990,14 +977,30 @@ static void macho_symdef(char *name, int32_t section, int64_t offset,
 {
     struct symbol *sym;
     struct section *s;
+    bool special_used = false;
 
-    if (special) {
-        nasm_error(ERR_NONFATAL, "The Mach-O output format does "
-              "not support any special symbol types");
-        return;
-    }
+#if defined(DEBUG) && DEBUG>2
+    nasm_error(ERR_DEBUG,
+            " macho_symdef: %s, pass=%"PRId64" type %s, sec=%"PRIx32", off=%"PRIx64", is_global=%d, %s\n",
+	       name, pass_count(), pass_types[pass_type()],
+	       section, offset, is_global, special);
+#endif
 
     if (is_global == 3) {
+        if (special) {
+            int n = strcspn(special, " \t");
+
+            if (!nasm_strnicmp(special, "private_extern", n)) {
+                for (sym = syms; sym != NULL; sym = sym->next) {
+                    if (!strcmp(name, sym->name)) {
+                        if (sym->type & N_PEXT)
+                            return; /* nothing to be done */
+                        else
+                            break;
+                    }
+                }
+            }
+        }
         nasm_error(ERR_NONFATAL, "The Mach-O format does not "
               "(yet) support forward reference fixups.");
         return;
@@ -1030,6 +1033,21 @@ static void macho_symdef(char *name, int32_t section, int64_t offset,
     if (is_global != 0) {
         sym->type |= N_EXT;
     }
+    if (is_global == 1) {
+        /* check special to see if the global symbol shall be marked as private external: N_PEXT */
+        if (special) {
+            int n = strcspn(special, " \t");
+
+            if (!nasm_strnicmp(special, "private_extern", n))
+                sym->type |= N_PEXT;
+            else
+                nasm_error(ERR_NONFATAL, "unrecognised symbol type `%.*s'", n, special);
+        }
+        special_used = true;
+    }
+
+    /* track the initially allocated symbol number for use in future fix-ups */
+    sym->initial_snum = nsyms;
 
     if (section == NO_SEG) {
         /* symbols in no section get absolute */
@@ -1045,9 +1063,6 @@ static void macho_symdef(char *name, int32_t section, int64_t offset,
         /* get the in-file index of the section the symbol was defined in */
         sym->sect = s ? s->fileindex : NO_SECT;
 
-	/* track the initially allocated symbol number for use in future fix-ups */
-	sym->initial_snum = nsyms;
-
         if (!s) {
             /* remember symbol number of references to external
              ** symbols, this works because every external symbol gets
@@ -1068,7 +1083,7 @@ static void macho_symdef(char *name, int32_t section, int64_t offset,
                 /* give an error on unfound section if it's not an
                  ** external or common symbol (assemble_file() does a
                  ** seg_alloc() on every call for them) */
-                nasm_panic(0, "in-file index for section %d not found, is_global = %d", section, is_global);
+                nasm_panic("in-file index for section %d not found, is_global = %d", section, is_global);
 		break;
             }
 	}
@@ -1081,6 +1096,9 @@ static void macho_symdef(char *name, int32_t section, int64_t offset,
     }
 
     ++nsyms;
+
+    if (special && !special_used)
+        nasm_error(ERR_NONFATAL, "no special symbol features supported here");
 }
 
 static void macho_sectalign(int32_t seg, unsigned int value)
@@ -1100,17 +1118,6 @@ static void macho_sectalign(int32_t seg, unsigned int value)
         s->align = align;
 }
 
-static int32_t macho_segbase(int32_t section)
-{
-    return section;
-}
-
-static void macho_filename(char *inname, char *outname)
-{
-    standard_extension(inname, outname, ".o");
-    module_name = inname;
-}
-
 extern macros_t macho_stdmac[];
 
 /* Comparison function for qsort symbol layout.  */
@@ -1269,7 +1276,7 @@ static void macho_calculate_sizes (void)
     }
 
     if (seg_nsects > MAX_SECT) {
-	nasm_fatal(0, "MachO output is limited to %d sections\n",
+	nasm_fatal("MachO output is limited to %d sections\n",
 		   MAX_SECT);
     }
 
@@ -1644,7 +1651,7 @@ static void macho_write (void)
     if (seg_nsects > 0)
 	offset = macho_write_segment (offset);
     else
-        nasm_error(ERR_WARNING, "no sections?");
+        nasm_warn(WARN_OTHER, "no sections?");
 
     if (nsyms > 0) {
         /* write out symbol command */
@@ -1711,6 +1718,7 @@ static void macho_cleanup(void)
     }
 
     saa_free(strs);
+
     raa_free(extsyms);
 
     while (syms) {
@@ -1722,6 +1730,8 @@ static void macho_cleanup(void)
     nasm_free(extdefsyms);
     nasm_free(undefsyms);
     nasm_free(sectstab);
+    raa_free(section_by_index);
+    hash_free(&section_by_name);
 }
 
 static bool macho_set_section_attribute_by_symbol(const char *label, uint32_t flags)
@@ -1730,7 +1740,7 @@ static bool macho_set_section_attribute_by_symbol(const char *label, uint32_t fl
     int32_t nasm_seg;
     int64_t offset;
 
-    if (!lookup_label(label, &nasm_seg, &offset)) {
+    if (lookup_label(label, &nasm_seg, &offset) == LBL_none) {
 	nasm_error(ERR_NONFATAL, "unknown symbol `%s' in no_dead_strip", label);
 	return false;
     }
@@ -1753,7 +1763,6 @@ static enum directive_result macho_no_dead_strip(const char *labels)
     char *s, *p, *ep;
     char ec;
     enum directive_result rv = DIRR_ERROR;
-    bool real = passn > 1;
 
     p = s = nasm_strdup(labels);
     while (*p) {
@@ -1768,7 +1777,7 @@ static enum directive_result macho_no_dead_strip(const char *labels)
 	    goto err;
 	}
 	*ep = '\0';
-	if (real) {
+	if (!pass_first()) {
 	    if (!macho_set_section_attribute_by_symbol(p, S_ATTR_NO_DEAD_STRIP))
 		rv = DIRR_ERROR;
 	}
@@ -1791,16 +1800,17 @@ err:
 static enum directive_result
 macho_pragma(const struct pragma *pragma)
 {
-    bool real = passn > 1;
-
     switch (pragma->opcode) {
     case D_SUBSECTIONS_VIA_SYMBOLS:
 	if (*pragma->tail)
 	    return DIRR_BADPARAM;
 
-	if (real)
+	if (!pass_first())
 	    head_flags |= MH_SUBSECTIONS_VIA_SYMBOLS;
 
+        /* Jmp-match optimization conflicts */
+        optimizing.flag |= OPTIM_DISABLE_JMP_MATCH;
+
 	return DIRR_OK;
 
     case D_NO_DEAD_STRIP:
@@ -1827,10 +1837,10 @@ static void macho_dbg_generate(void)
     /* debug section defines */
     {
         int bits = 0;
-        macho_section(".debug_abbrev", 0, &bits);
-        macho_section(".debug_info", 0, &bits);
-        macho_section(".debug_line", 0, &bits);
-        macho_section(".debug_str", 0, &bits);
+        macho_section(".debug_abbrev", &bits);
+        macho_section(".debug_info", &bits);
+        macho_section(".debug_line", &bits);
+        macho_section(".debug_str", &bits);
     }
 
     /* dw section walk to find high_addr and total_len */
@@ -1932,7 +1942,7 @@ static void macho_dbg_generate(void)
         nasm_assert(p_section != NULL);
 
         producer_str_offset = 0;
-        module_str_offset = dir_str_offset = saa_wcstring(p_str, nasm_signature);
+        module_str_offset = dir_str_offset = saa_wcstring(p_str, nasm_signature());
         dir_str_offset += saa_wcstring(p_str, cur_file);
         saa_wcstring(p_str, cur_dir);
 
@@ -2274,7 +2284,7 @@ static void macho32_init(void)
 }
 
 static const struct dfmt macho32_df_dwarf = {
-    "MachO32 (i386) dwarf debug format for Darwin/MacOS",
+    "Mach-O i386 dwarf for Darwin/MacOS",
     "dwarf",
     macho_dbg_init,
     macho_dbg_linenum,
@@ -2290,22 +2300,24 @@ static const struct dfmt * const macho32_df_arr[2] =
  { &macho32_df_dwarf, NULL };
 
 const struct ofmt of_macho32 = {
-    "NeXTstep/OpenStep/Rhapsody/Darwin/MacOS X (i386) object files",
+    "Mach-O i386 (Mach, including MacOS X and variants)",
     "macho32",
+    ".o",
     0,
     32,
     macho32_df_arr,
     &macho32_df_dwarf,
     macho_stdmac,
     macho32_init,
+    null_reset,
     nasm_do_legacy_output,
     macho_output,
     macho_symdef,
     macho_section,
+    macho_herelabel,
     macho_sectalign,
-    macho_segbase,
+    null_segbase,
     null_directive,
-    macho_filename,
     macho_cleanup,
     macho_pragma_list
 };
@@ -2335,11 +2347,11 @@ static void macho64_init(void)
 
     /* add special symbol for ..gotpcrel */
     macho_gotpcrel_sect = seg_alloc() + 1;
-    define_label("..gotpcrel", macho_gotpcrel_sect, 0L, NULL, false, false);
+    backend_label("..gotpcrel", macho_gotpcrel_sect, 0L);
 }
 
 static const struct dfmt macho64_df_dwarf = {
-    "MachO64 (x86-64) dwarf debug format for Darwin/MacOS",
+    "Mach-O x86-64 dwarf for Darwin/MacOS",
     "dwarf",
     macho_dbg_init,
     macho_dbg_linenum,
@@ -2355,22 +2367,24 @@ static const struct dfmt * const macho64_df_arr[2] =
  { &macho64_df_dwarf, NULL };
 
 const struct ofmt of_macho64 = {
-    "NeXTstep/OpenStep/Rhapsody/Darwin/MacOS X (x86_64) object files",
+    "Mach-O x86-64 (Mach, including MacOS X and variants)",
     "macho64",
+    ".o",
     0,
     64,
     macho64_df_arr,
     &macho64_df_dwarf,
     macho_stdmac,
     macho64_init,
+    null_reset,
     nasm_do_legacy_output,
     macho_output,
     macho_symdef,
     macho_section,
+    macho_herelabel,
     macho_sectalign,
-    macho_segbase,
+    null_segbase,
     null_directive,
-    macho_filename,
     macho_cleanup,
     macho_pragma_list,
 };
