@@ -6,9 +6,6 @@
 
 #include "compiler.h"
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <stdarg.h>
 
 #include "nasmlib.h"
 
